import datetime as dt

from django.test import TestCase
from django.utils.timezone import now

from structures.helpers import (
    datetime_almost_equal,
    hours_until_deadline,
    is_absolute_url,
)


class TestDatetimeAlmostEqual(TestCase):
    def test_should_return_true(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        d2 = now() + dt.timedelta(hours=1, minutes=5)
        # when / then
        self.assertTrue(datetime_almost_equal(d1, d2, 3600))
        self.assertTrue(datetime_almost_equal(d2, d1, 3600))
        self.assertFalse(datetime_almost_equal(d2, d1, 60))
        self.assertFalse(datetime_almost_equal(d1, d2, 60))

    def test_should_return_false(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        d2 = now() + dt.timedelta(hours=1, minutes=5)
        # when / then
        self.assertFalse(datetime_almost_equal(d2, d1, 60))
        self.assertFalse(datetime_almost_equal(d1, d2, 60))

    def test_should_return_false_for_none_dates(self):
        # given
        d1 = now() + dt.timedelta(hours=0, minutes=55)
        # when / then
        self.assertFalse(datetime_almost_equal(d1, None, 3600))
        self.assertFalse(datetime_almost_equal(None, d1, 3600))


class TestHoursUntilDeadline(TestCase):
    def test_should_return_correct_value_for_two_datetimes(self):
        # given
        d1 = now()
        d2 = d1 - dt.timedelta(hours=3)
        # when / then
        self.assertEqual(hours_until_deadline(d1, d2), 3)

    def test_should_return_correct_value_for_one_datetimes(self):
        # given
        d1 = now() + dt.timedelta(hours=3)
        # when / then
        self.assertAlmostEqual(hours_until_deadline(d1), 3, delta=0.1)

    def test_should_raise_error_when_deadline_is_not_a_datetime(self):
        with self.assertRaises(TypeError):
            hours_until_deadline(None)


class TestIsAbsoluteUrl(TestCase):
    def test_should_detect_absolute_urls(self):
        cases = [
            ("https://www.google.com", True),
            ("http://www.google.com", True),
            ("www.google.com", False),
            ("", False),
            ("/abc/x", False),
            (None, False),
        ]
        for url, expected_result in cases:
            with self.subTest(url=url):
                self.assertIs(is_absolute_url(url), expected_result)



import datetime as dt
from unittest.mock import patch

from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveCorporationInfo

from structures.admin import (
    NotificationAdmin,
    OwnerAdmin,
    OwnerAllianceFilter,
    OwnerCorporationsFilter,
    StructureAdmin,
    StructureFuelAlertConfigAdmin,
    WebhookAdmin,
)
from structures.models import (
    FuelAlertConfig,
    Notification,
    Owner,
    Structure,
    StructureTag,
    Webhook,
)

from .testdata.factories_2 import (
    FuelAlertConfigFactory,
    NotificationFactory,
    OwnerFactory,
    StructureFactory,
)
from .testdata.helpers import (
    create_structures,
    create_user,
    load_entities,
    load_notification_entities,
    set_owner_character,
)
from .testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.admin"


class MockRequest(object):
    def __init__(self, user=None):
        self.user = user


class TestFuelNotificationConfigAdminView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.defaults = {
            "is_enabled": True,
            "channel_ping_type": Webhook.PingType.HERE,
            "color": Webhook.Color.WARNING,
        }
        cls.user = User.objects.create_superuser("Clark Kent")
        load_eveuniverse()

    def test_should_create_new_config(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 12, "end": 5, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_update_existing_config(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={**self.defaults, **{"start": 48, "end": 0, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_remove_existing_fuel_notifications_when_timing_changed(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        structure = StructureFactory()
        structure.structure_fuel_alerts.create(
            config=config, structure=structure, hours=5
        )
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={**self.defaults, **{"start": 48, "end": 0, "repeat": 2}},
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    def test_should_not_remove_existing_fuel_notifications_on_other_changes(self):
        # given
        self.client.force_login(self.user)
        config = FuelAlertConfigFactory(start=48, end=24, repeat=12)
        structure = StructureFactory()
        structure.structure_fuel_alerts.create(
            config=config, structure=structure, hours=5
        )
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_change", args=[config.pk]),
            data={
                **self.defaults,
                **{"start": 48, "end": 24, "repeat": 12, "is_enabled": False},
            },
        )
        # then
        self.assertRedirects(
            response, reverse("admin:structures_fuelalertconfig_changelist")
        )
        self.assertEqual(structure.structure_fuel_alerts.count(), 1)

    def test_should_not_allow_end_before_start(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 1, "end": 2, "repeat": 1}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_not_allow_invalid_frequency(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 48, "end": 24, "repeat": 36}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_not_allow_creating_overlapping(self):
        # given
        self.client.force_login(self.user)
        FuelAlertConfigFactory(start=48, end=24, repeat=12)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 0, "repeat": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 1)

    def test_should_work_with_empty_end_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "repeat": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_work_with_empty_start_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)

    def test_should_work_with_empty_repeat_field(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.post(
            reverse("admin:structures_fuelalertconfig_add"),
            data={**self.defaults, **{"start": 36, "end": 8}},
        )
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "errornote")
        self.assertEqual(FuelAlertConfig.objects.count(), 0)


class TestStructureFuelAlertAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = StructureFuelAlertConfigAdmin(
            model=FuelAlertConfig, admin_site=AdminSite()
        )
        load_eveuniverse()

    @patch(MODULE_PATH + ".StructureFuelAlertConfigAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".tasks", spec=True)
    def test_should_send_fuel_notifications(self, mock_tasks, mock_message_user):
        # given
        config = FuelAlertConfigFactory()
        request = MockRequest()
        queryset = FuelAlertConfig.objects.filter(pk=config.pk)
        # when
        self.modeladmin.send_fuel_notifications(request, queryset)
        # then
        self.assertTrue(mock_tasks.send_queued_messages_for_webhooks.called)
        self.assertTrue(mock_message_user.called)


class TestNotificationAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = NotificationAdmin(model=Notification, admin_site=AdminSite())
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        load_notification_entities(cls.owner)
        cls.obj = Notification.objects.get(notification_id=1000000404)
        cls.obj_qs = Notification.objects.filter(
            notification_id__in=[1000000404, 1000000405]
        )

    def test_structures_1(self):
        # given
        notif = Notification.objects.get(notification_id=1000000404)
        # when
        result = self.modeladmin._structures(notif)
        self.assertEqual(result, "?")

    def test_structures_2(self):
        # given
        notif = Notification.objects.get(notification_id=1000000202)
        # when
        result = self.modeladmin._structures(notif)
        self.assertIsNone(result)

    # FixMe: Does not seam to work with special prefetch list
    # def test_webhooks(self):
    #     self.owner.webhooks.add(Webhook.objects.get(name="Test Webhook 2"))
    #     self.assertEqual(
    #         self.modeladmin._webhooks(self.obj), "Test Webhook 1, Test Webhook 2"
    #     )

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", autospec=True)
    def test_action_mark_as_sent(self, mock_message_user):
        for obj in self.obj_qs:
            obj.is_sent = False
            obj.save()
        self.modeladmin.mark_as_sent(MockRequest(self.user), self.obj_qs)
        for obj in self.obj_qs:
            self.assertTrue(obj.is_sent)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", autospec=True)
    def test_action_mark_as_unsent(self, mock_message_user):
        for obj in self.obj_qs:
            obj.is_sent = True
            obj.save()
        self.modeladmin.mark_as_unsent(MockRequest(self.user), self.obj_qs)
        for obj in self.obj_qs:
            self.assertFalse(obj.is_sent)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", autospec=True)
    @patch(MODULE_PATH + ".tasks.send_queued_messages_for_webhooks")
    def test_action_send_to_webhook(self, mock_task, mock_message_user):
        self.modeladmin.send_to_configured_webhooks(MockRequest(self.user), self.obj_qs)
        self.assertEqual(mock_task.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".NotificationAdmin.message_user", autospec=True)
    @patch(MODULE_PATH + ".Notification.add_or_remove_timer")
    def test_action_process_for_timerboard(
        self, mock_process_for_timerboard, mock_message_user
    ):
        self.modeladmin.add_or_remove_timer(MockRequest(self.user), self.obj_qs)
        self.assertEqual(mock_process_for_timerboard.call_count, 2)
        self.assertTrue(mock_message_user.called)


class TestOwnerAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.modeladmin = OwnerAdmin(model=Owner, admin_site=AdminSite())
        load_eveuniverse()
        create_structures()
        cls.user, cls.obj = set_owner_character(character_id=1001)

    def test_corporation(self):
        self.assertEqual(self.modeladmin._corporation(self.obj), "Wayne Technologies")

    def test_alliance_normal(self):
        self.assertEqual(self.modeladmin._alliance(self.obj), "Wayne Enterprises")

    def test_alliance_none(self):
        my_owner = Owner.objects.get(corporation__corporation_id=2102)
        self.assertIsNone(self.modeladmin._alliance(my_owner))

    def test_webhooks(self):
        self.obj.webhooks.add(Webhook.objects.get(name="Test Webhook 2"))
        self.assertEqual(
            self.modeladmin._webhooks(self.obj), "Test Webhook 1<br>Test Webhook 2"
        )

    @patch(MODULE_PATH + ".OwnerAdmin.message_user", autospec=True)
    @patch(MODULE_PATH + ".tasks.update_structures_for_owner")
    def test_action_update_structures(self, mock_task, mock_message_user):
        owner_qs = Owner.objects.filter(corporation__corporation_id__in=[2001, 2002])
        self.modeladmin.update_structures(MockRequest(self.user), owner_qs)
        self.assertEqual(mock_task.delay.call_count, 2)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".OwnerAdmin.message_user", autospec=True)
    @patch(MODULE_PATH + ".tasks.process_notifications_for_owner")
    def test_action_fetch_notifications(self, mock_task, mock_message_user):
        owner_qs = Owner.objects.filter(corporation__corporation_id__in=[2001, 2002])
        self.modeladmin.fetch_notifications(MockRequest(self.user), owner_qs)
        self.assertEqual(mock_task.delay.call_count, 2)
        self.assertTrue(mock_message_user.called)

    def test_should_return_empty_turnaround_times(self):
        # given
        my_owner = Owner.objects.get(corporation__corporation_id=2001)
        # when
        result = self.modeladmin._avg_turnaround_time(my_owner)
        # then
        self.assertEqual(result, "- | - | -")

    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_SHORT", 5)
    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MEDIUM", 15)
    @patch(MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_LONG", 50)
    @patch(
        MODULE_PATH + ".app_settings.STRUCTURES_NOTIFICATION_TURNAROUND_MAX_VALID", 3600
    )
    def test_should_return_correct_turnaround_times(self):
        # given
        my_owner = OwnerFactory()
        my_sender = EveEntity.objects.get(id=1001)
        my_now = now()
        NotificationFactory(
            owner=my_owner,
            notification_id=1,
            sender=my_sender,
            last_updated=my_now,
            timestamp=my_now,
            created=my_now + dt.timedelta(seconds=3601),
        )
        for i in range(50):
            timestamp = my_now + dt.timedelta(minutes=i)
            NotificationFactory(
                owner=my_owner,
                notification_id=2 + i,
                sender=my_sender,
                last_updated=my_now,
                timestamp=timestamp,
                created=timestamp + dt.timedelta(seconds=2),
            )
        # when
        result = self.modeladmin._avg_turnaround_time(my_owner)
        # then
        self.assertEqual(result, "2 | 2 | 2")


class TestStructureAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.modeladmin = StructureAdmin(model=Structure, admin_site=AdminSite())
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        cls.obj = Structure.objects.get(id=1000000000001)
        cls.obj_qs = Structure.objects.filter(id__in=[1000000000001, 1000000000002])

    def test_owner(self):
        self.assertEqual(
            self.modeladmin._owner(self.obj), "Wayne Technologies<br>Wayne Enterprises"
        )

    def test_location(self):
        self.assertEqual(self.modeladmin._location(self.obj), "Amamake<br>Heimatar")
        poco = Structure.objects.get(id=1200000000003)
        self.assertEqual(self.modeladmin._location(poco), "Amamake V<br>Heimatar")
        starbase = Structure.objects.get(id=1300000000001)
        self.assertEqual(
            self.modeladmin._location(starbase), "Amamake II - Moon 1<br>Heimatar"
        )

    def test_type(self):
        self.assertEqual(self.modeladmin._type(self.obj), "Astrahus<br>Citadel")

    def test_tags_1(self):
        self.assertSetEqual(set(self.modeladmin._tags(self.obj)), {"lowsec", "tag_a"})

    def test_tags_2(self):
        self.obj.tags.clear()
        self.assertListEqual(self.modeladmin._tags(self.obj), [])

    @patch(MODULE_PATH + ".StructureAdmin.message_user", autospec=True)
    def test_action_add_default_tags(self, mock_message_user):
        for obj in self.obj_qs:
            obj.tags.clear()
        self.modeladmin.add_default_tags(MockRequest(self.user), self.obj_qs)
        default_tags = StructureTag.objects.filter(is_default=True)
        for obj in self.obj_qs:
            self.assertSetEqual(set(obj.tags.all()), set(default_tags))
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".StructureAdmin.message_user", autospec=True)
    def test_action_remove_user_tags(self, mock_message_user):
        self.modeladmin.remove_user_tags(MockRequest(self.user), self.obj_qs)
        for obj in self.obj_qs:
            self.assertFalse(obj.tags.filter(is_user_managed=True).exists())
        self.assertTrue(mock_message_user.called)

    def test_owner_corporations_status_filter(self):
        class StructureAdminTest(admin.ModelAdmin):
            list_filter = (OwnerCorporationsFilter,)

        Owner.objects.all().delete()
        owner_2001 = OwnerFactory(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )
        OwnerFactory(corporation=EveCorporationInfo.objects.get(corporation_id=2002))
        my_modeladmin = StructureAdminTest(Structure, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get("/")
        request.user = self.user
        changelist = my_modeladmin.get_changelist_instance(request)
        filterspec = changelist.get_filters(request)[0][0]
        expected = [(2002, "Wayne Foods"), (2001, "Wayne Technologies")]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get("/", {"owner_corporation_id__exact": 2001})
        request.user = self.user
        changelist = my_modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Structure.objects.filter(owner=owner_2001)
        self.assertSetEqual(set(queryset), set(expected))

    def test_owner_alliance_status_filter(self):
        class StructureAdminTest(admin.ModelAdmin):
            list_filter = (OwnerAllianceFilter,)

        Owner.objects.all().delete()
        owner_2001 = OwnerFactory(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )
        owner_2002 = OwnerFactory(
            corporation=EveCorporationInfo.objects.get(corporation_id=2002)
        )
        OwnerFactory(corporation=EveCorporationInfo.objects.get(corporation_id=2102))
        modeladmin = StructureAdminTest(Structure, AdminSite())

        # Make sure the lookups are correct
        request = self.factory.get("/")
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        filterspec = changelist.get_filters(request)[0][0]
        expected = [(3001, "Wayne Enterprises")]
        self.assertEqual(filterspec.lookup_choices, expected)

        # Make sure the correct queryset is returned
        request = self.factory.get("/", {"owner_alliance_id__exact": 3001})
        request.user = self.user
        changelist = modeladmin.get_changelist_instance(request)
        queryset = changelist.get_queryset(request)
        expected = Structure.objects.filter(owner__in=[owner_2001, owner_2002])
        self.assertSetEqual(set(queryset), set(expected))


class TestWebhookAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = WebhookAdmin(model=Webhook, admin_site=AdminSite())
        load_entities([Webhook])
        cls.user = create_user(character_id=1001, load_data=True)
        cls.obj_qs = Webhook.objects.all()

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", autospec=True)
    @patch(MODULE_PATH + ".tasks.send_test_notifications_to_webhook")
    def test_action_test_notification(self, mock_task, mock_message_user):
        self.modeladmin.test_notification(MockRequest(self.user), self.obj_qs)
        self.assertEqual(mock_task.delay.call_count, 2)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", autospec=True)
    def test_action_activate(self, mock_message_user):
        for obj in self.obj_qs:
            obj.is_active = False
            obj.save()
        self.modeladmin.activate(MockRequest(self.user), self.obj_qs)
        for obj in self.obj_qs:
            self.assertTrue(obj.is_active)

        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".WebhookAdmin.message_user", autospec=True)
    def test_action_deactivate(self, mock_message_user):
        for obj in self.obj_qs:
            obj.is_active = True
            obj.save()
        self.modeladmin.deactivate(MockRequest(self.user), self.obj_qs)
        for obj in self.obj_qs:
            self.assertFalse(obj.is_active)

        self.assertTrue(mock_message_user.called)



# flake8: noqa
"""scripts generates large amount of random structures for load testing"""

import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(currentdir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)

from datetime import timedelta
from random import randrange

import django
from django.apps import apps
from django.db import transaction
from django.utils.timezone import now

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

if not apps.is_installed("structures"):
    raise RuntimeError("The app structures is not installed")

from esi.clients import esi_client_factory

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from structures.models import (
    EveSolarSystem,
    EveType,
    Owner,
    Structure,
    StructureService,
    StructureTag,
)

# TODO: Add data for assets, e.g. fittings

print(
    "generate_structure - "
    "scripts generates large amount of random structures for load testing "
)

amount = 20

# random pick of most active corporations on zKillboard in Jan 2020
corporation_ids = [
    98388312,
    98558506,
    98370861,
    98410772,
    98148549,
    98431483,
    667531913,
    427125032,
    98514543,
    98013740,
]
structure_type_ids = [35825, 35826, 35827, 35832, 35832, 35834, 35835, 35836]
solar_system_ids = [30000142, 30001445, 30002355, 30004046, 30003833, 30045338]
services = [
    "Clone Bay",
    "Moondrilling",
    "Reprocessing",
    "Market Hub",
    "Manufacturing (Standard)",
    "Blueprint Copying",
    "Material Efficiency Research",
    "Time Efficiency Research",
]
tag_names = [
    "Top Secret",
    "Priority",
    "Trash",
    "Needs caretaker",
    "Taskforce Bravo",
    "Not so friendly",
]


def get_random(lst: list) -> object:
    return lst[randrange(len(lst))]


def get_random_subset(lst: list, max_members: int = None) -> list:
    lst2 = lst.copy()
    subset = list()
    if not max_members:
        max_members = len(lst)
    else:
        max_members = min(max_members, len(lst))

    for x in range(randrange(max_members) + 1):
        m = lst2.pop(randrange(len(lst2)))
        subset.append(m)

    return subset


print("Connecting to ESI ...")
client = esi_client_factory()

# generating data
print("Creating base data ...")
owners = list()
for corporation_id in corporation_ids:
    try:
        corporation = client.Corporation.get_corporations_corporation_id(
            corporation_id=corporation_id
        ).result()
        try:
            EveAllianceInfo.objects.get(alliance_id=corporation["alliance_id"])
        except EveAllianceInfo.DoesNotExist:
            EveAllianceInfo.objects.create_alliance(corporation["alliance_id"])
    except Exception:
        pass

    try:
        corporation = EveCorporationInfo.objects.get(corporation_id=corporation_id)
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(corporation_id)

    owner, _ = Owner.objects.get_or_create(corporation=corporation)
    owners.append(owner)

eve_types = list()
for type_id in structure_type_ids:
    eve_type, _ = EveType.objects.get_or_create_esi(id=type_id)
    eve_types.append(eve_type)

eve_solar_systems = list()
for system_id in solar_system_ids:
    eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(id=system_id)
    eve_solar_systems.append(eve_solar_system)

tags = list()
for name in tag_names:
    tag, _ = StructureTag.objects.update_or_create(
        name=name,
        defaults={"style": get_random([x[0] for x in StructureTag.Style.choices])},
    )
    tags.append(tag)

# creating structures
print("Creating {} structures ...".format(amount))
Structure.objects.filter(owner__in=owners).delete()
with transaction.atomic():
    for i in range(1, amount + 1):
        state = get_random(
            [
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.SHIELD_VULNERABLE,
                Structure.State.ARMOR_REINFORCE,
                Structure.State.HULL_REINFORCE,
            ]
        )
        is_low_power = (
            get_random([True, False]) or state == Structure.State.HULL_REINFORCE
        )

        unanchors_at = None

        if not is_low_power:
            fuel_expires_at = now() + timedelta(days=randrange(14), hours=randrange(12))
            last_online_at = now()
            if randrange(1, 10) == 1:
                unanchors_at = now() + timedelta(days=randrange(6), hours=randrange(12))
        else:
            fuel_expires_at = None
            last_online_at = now() - timedelta(days=get_random([1, 2, 3, 10]))
        structure = Structure.objects.create(
            id=1000000000001 + i,
            owner=get_random(owners),
            eve_type=get_random(eve_types),
            name="Generated structure #{:05d}".format(i),
            eve_solar_system=get_random(eve_solar_systems),
            reinforce_hour=randrange(24),
            state=state,
            fuel_expires_at=fuel_expires_at,
            last_online_at=last_online_at,
            unanchors_at=unanchors_at,
        )
        if is_low_power:
            state = StructureService.State.OFFLINE
        else:
            state = StructureService.State.ONLINE
        for name in get_random_subset(services, 3):
            StructureService.objects.create(structure=structure, name=name, state=state)
        structure.tags.add(*get_random_subset(tags))
        if structure.is_reinforced:
            state_timer_start = now() - timedelta(
                days=randrange(3), hours=randrange(12)
            )
            state_timer_end = now() + timedelta(days=randrange(3), hours=randrange(12))
            structure.state_timer_start = state_timer_start
            structure.state_timer_end = state_timer_end

        structure.save()

print("DONE")



"""Helpers for Structures."""

import datetime as dt
from typing import Optional
from urllib.parse import urlparse

from django.utils.timezone import now


def hours_until_deadline(
    deadline: dt.datetime, start: Optional[dt.datetime] = None
) -> float:
    """Currently remaining hours until a given deadline."""
    if not isinstance(deadline, dt.datetime):
        raise TypeError("deadline must be of type datetime")
    if not start:
        start = now()
    return (deadline - start).total_seconds() / 3600


def datetime_almost_equal(
    first: Optional[dt.datetime], second: Optional[dt.datetime], threshold: int
) -> bool:
    """True when first and second datetime are within threshold in seconds.
    False when first or second is None.
    """
    if not first or not second:
        return False
    dif = abs((first - second).total_seconds())
    return dif <= abs(threshold)


def is_absolute_url(url: str) -> bool:
    """Return True if URL is absolute else False."""
    return bool(urlparse(url).netloc)



from app_utils.django import app_labels

if "structuretimers" in app_labels():
    import datetime as dt
    from unittest.mock import Mock, patch

    import pytz
    from structuretimers.models import Timer

    from django.utils.timezone import now
    from eveuniverse.models import EveSolarSystem, EveType

    from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
    from app_utils.testing import NoSocketsTestCase

    from structures.core import notification_timers
    from structures.models import Notification
    from structures.tests.testdata.factories import create_webhook
    from structures.tests.testdata.factories_2 import GeneratedNotificationFactory
    from structures.tests.testdata.helpers import (
        create_structures,
        load_notification_entities,
        set_owner_character,
    )
    from structures.tests.testdata.load_eveuniverse import load_eveuniverse

    MODULE_PATH = "structures.core.notification_timers"

    @patch(
        "structuretimers.models._task_calc_timer_distances_for_all_staging_systems",
        Mock(),
    )
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
    class TestTimersForStructureTimers(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            load_eveuniverse()
            create_structures()

        def setUp(self) -> None:
            _, self.owner = set_owner_character(character_id=1001)
            load_notification_entities(self.owner)
            self.owner.webhooks.add(create_webhook())

        def test_should_create_timer_for_reinforced_structure(self):
            # given
            notif = Notification.objects.get(notification_id=1000000505)
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(
                timer.eve_solar_system, EveSolarSystem.objects.get(id=30002537)
            )
            self.assertEqual(timer.structure_type, EveType.objects.get(id=35832))
            self.assertEqual(timer.timer_type, Timer.Type.ARMOR)
            self.assertEqual(timer.objective, Timer.Objective.FRIENDLY)
            self.assertAlmostEqual(
                timer.date, now() + dt.timedelta(hours=47), delta=dt.timedelta(hours=1)
            )
            self.assertEqual(
                timer.eve_corporation,
                EveCorporationInfo.objects.get(corporation_id=2001),
            )
            self.assertEqual(
                timer.eve_alliance, EveAllianceInfo.objects.get(alliance_id=3001)
            )
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.structure_name, "Test Structure Alpha")
            self.assertEqual(timer.owner_name, "Wayne Technologies")
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_sov_reinforcement(self):
            # given
            notif = Notification.objects.get(notification_id=1000000804)
            self.owner.is_alliance_main = True
            self.owner.save()
            notif.owner.refresh_from_db()
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.FINAL)
            self.assertEqual(
                timer.eve_solar_system, EveSolarSystem.objects.get(id=30000474)
            )
            self.assertEqual(timer.structure_type, EveType.objects.get(id=32226))
            self.assertAlmostEqual(
                timer.date,
                pytz.utc.localize(dt.datetime(2018, 12, 20, 17, 3, 22)),
                delta=dt.timedelta(seconds=120),
            )
            self.assertEqual(
                timer.eve_corporation,
                EveCorporationInfo.objects.get(corporation_id=2001),
            )
            self.assertEqual(
                timer.eve_alliance, EveAllianceInfo.objects.get(alliance_id=3001)
            )
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.owner_name, "Wayne Enterprises")
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_sov_reinforcement_2(self):
            # given
            notif = Notification.objects.get(notification_id=1000000804)
            self.owner.is_alliance_main = False
            self.owner.save()
            notif.owner.refresh_from_db()
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertFalse(result)
            self.assertFalse(Timer.objects.exists())
            notif.refresh_from_db()
            self.assertFalse(notif.is_timer_added)

        def test_should_create_timer_for_orbital_reinforcements(self):
            # given
            notif = Notification.objects.get(notification_id=1000000602)
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.FINAL)
            self.assertEqual(
                timer.eve_solar_system, EveSolarSystem.objects.get(id=30002537)
            )
            self.assertEqual(timer.structure_type, EveType.objects.get(id=2233))
            self.assertEqual(timer.location_details, "Amamake IV")
            self.assertAlmostEqual(
                timer.date,
                pytz.utc.localize(dt.datetime(2019, 10, 13, 20, 32, 27)),
                delta=dt.timedelta(seconds=120),
            )
            self.assertEqual(
                timer.eve_corporation,
                EveCorporationInfo.objects.get(corporation_id=2001),
            )
            self.assertEqual(
                timer.eve_alliance, EveAllianceInfo.objects.get(alliance_id=3001)
            )
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.owner_name, "Wayne Technologies")
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_moon_extraction(self):
            # given
            notif = Notification.objects.get(notification_id=1000000404)
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            self.assertEqual(timer.timer_type, Timer.Type.MOONMINING)
            self.assertEqual(
                timer.eve_solar_system, EveSolarSystem.objects.get(id=30002537)
            )
            self.assertEqual(timer.structure_type, EveType.objects.get(id=35835))
            self.assertEqual(
                timer.eve_corporation,
                EveCorporationInfo.objects.get(corporation_id=2001),
            )
            self.assertEqual(
                timer.eve_alliance, EveAllianceInfo.objects.get(alliance_id=3001)
            )
            self.assertEqual(timer.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(timer.location_details, "Amamake II - Moon 1")
            self.assertEqual(timer.owner_name, "Wayne Technologies")
            self.assertEqual(timer.structure_name, "Dummy")
            self.assertTrue(timer.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_raise_error_for_unsupported_types(self):
            notif = Notification.objects.get(notification_id=1000010501)
            # when
            with self.assertRaises(NotImplementedError):
                notification_timers.add_or_remove_timer(notif)

        def test_can_delete_extraction_timer(self):
            # create timer
            notif = Notification.objects.get(notification_id=1000000404)
            self.assertTrue(notification_timers.add_or_remove_timer(notif))
            timer = Timer.objects.first()
            self.assertIsInstance(timer, Timer)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

            # delete timer
            notif = Notification.objects.get(notification_id=1000000402)
            self.assertTrue(notification_timers.add_or_remove_timer(notif))
            self.assertFalse(Timer.objects.filter(pk=timer.pk).exists())
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)

        def test_should_create_timer_for_starbase_reinforcement(self):
            # given
            notif = GeneratedNotificationFactory()
            structure = notif.structures.first()
            # when
            result = notification_timers.add_or_remove_timer(notif)
            # then
            self.assertTrue(result)
            obj = Timer.objects.first()
            self.assertEqual(
                obj.eve_solar_system,
                EveSolarSystem.objects.get(id=structure.eve_solar_system.id),
            )
            self.assertEqual(
                obj.structure_type, EveType.objects.get(id=structure.eve_type.id)
            )
            self.assertEqual(obj.timer_type, Timer.Type.FINAL)
            self.assertEqual(obj.objective, Timer.Objective.FRIENDLY)
            self.assertAlmostEqual(obj.date, structure.state_timer_end)
            self.assertEqual(obj.eve_corporation, structure.owner.corporation)
            self.assertEqual(obj.visibility, Timer.Visibility.UNRESTRICTED)
            self.assertEqual(obj.structure_name, structure.name)
            self.assertEqual(
                obj.owner_name, structure.owner.corporation.corporation_name
            )
            self.assertTrue(obj.details_notes)
            notif.refresh_from_db()
            self.assertTrue(notif.is_timer_added)



import datetime as dt

import dhooks_lite

from django.test import TestCase, override_settings
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from structures.core.notification_embeds import NotificationBaseEmbed
from structures.core.notification_embeds.billing_embeds import BillType
from structures.core.notification_embeds.moonmining_embeds import (
    NotificationMoonminningExtractionFinished,
)
from structures.core.notification_embeds.tower_embeds import (
    NotificationTowerReinforcedExtra,
)
from structures.core.notification_types import NotificationType
from structures.models.notifications import Notification, Structure, Webhook
from structures.tests.testdata.factories import (
    create_notification,
    create_owner_from_user,
    create_starbase,
)
from structures.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    GeneratedNotificationFactory,
    NotificationFactory,
    OwnerFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import (
    create_structures,
    load_entities,
    load_notification_entities,
    markdown_to_plain,
    set_owner_character,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.core.notification_embeds"


class TestBilType(TestCase):
    def test_should_create_from_valid_id(self):
        self.assertEqual(BillType.to_enum(7), BillType.INFRASTRUCTURE_HUB)

    def test_should_create_from_invalid_id(self):
        for bill_id in range(7):
            with self.subTest(bill_id=bill_id):
                self.assertEqual(BillType.to_enum(bill_id), BillType.UNKNOWN)


class TestNotificationEmbeds(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        _, cls.owner = set_owner_character(character_id=1001)
        load_notification_entities(cls.owner)

    def test_should_create_obj_from_notification(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        # when
        notification_embed = NotificationBaseEmbed.create(notification)
        # then
        self.assertIsInstance(
            notification_embed, NotificationMoonminningExtractionFinished
        )
        self.assertEqual(
            str(notification_embed), "1000000403:MoonminingExtractionFinished"
        )
        self.assertEqual(
            repr(notification_embed),
            "NotificationMoonminningExtractionFinished(notification=Notification("
            "notification_id=1000000403, owner='Wayne Technologies', "
            "notif_type='MoonminingExtractionFinished'))",
        )

    def test_should_raise_exception_for_unsupported_notif_types(self):
        # given
        notification = Notification.objects.create(
            notification_id=666,
            owner=self.owner,
            sender=EveEntity.objects.get(id=2001),
            timestamp=now(),
            notif_type="XXXUnsupportedNotificationTypeXXX",
            last_updated=now(),
        )
        # when / then
        with self.assertRaises(NotImplementedError):
            NotificationBaseEmbed.create(notification)

    def test_should_require_notification_for_init(self):
        with self.assertRaises(TypeError):
            NotificationBaseEmbed(notification="dummy")

    def test_should_require_notification_for_factory(self):
        with self.assertRaises(TypeError):
            NotificationBaseEmbed.create(notification="dummy")


class TestNotificationEmbedsGenerate(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        _, cls.owner = set_owner_character(character_id=1001)
        load_notification_entities(cls.owner)
        cls.webhook = WebhookFactory()
        cls.owner.webhooks.add(cls.webhook)

    def test_should_generate_embed_from_notification(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertEqual(discord_embed.footer.text, "Eve Online")
        self.assertIn("eve_symbol_128.png", discord_embed.footer.icon_url)

    def test_should_generate_embed_from_notification_with_custom_color(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification._color_override = Webhook.Color.SUCCESS
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertEqual(discord_embed.color, Webhook.Color.SUCCESS)

    def test_should_generate_embed_from_notification_without_custom_color(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertIsInstance(discord_embed, dhooks_lite.Embed)
        self.assertNotEqual(discord_embed.color, Webhook.Color.SUCCESS)

    def test_should_generate_embed_from_notification_with_ping_type_override(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification._ping_type_override = Webhook.PingType.EVERYONE
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.EVERYONE)

    def test_should_generate_embed_for_all_supported_esi_notification_types(self):
        for notif_type in NotificationType.esi_notifications():
            with self.subTest(notif_type=notif_type):
                # given
                notification = (
                    Notification.objects.select_related("owner", "sender")
                    .filter(notif_type=notif_type)
                    .first()
                )
                notification_embed = NotificationBaseEmbed.create(notification)
                # when
                discord_embed = notification_embed.generate_embed()
                # then
                self.assertIsInstance(discord_embed, dhooks_lite.Embed)

    def test_should_set_ping_everyone_for_color_danger(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.DANGER
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.EVERYONE)

    def test_should_set_ping_everyone_for_color_warning(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.WARNING
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.HERE)

    def test_should_not_set_ping_everyone_for_color_info(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.INFO
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.NONE)

    def test_should_not_set_ping_everyone_for_color_success(self):
        # given
        notification = Notification.objects.get(notification_id=1000000513)
        notification_embed = NotificationBaseEmbed.create(notification)
        notification_embed._color = Webhook.Color.SUCCESS
        # when
        notification_embed.generate_embed()
        # then
        self.assertEqual(notification_embed.ping_type, Webhook.PingType.NONE)

    @override_settings(DEBUG=True)
    def test_should_set_footer_in_developer_mode(self):
        # given
        notification = Notification.objects.get(notification_id=1000000403)
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertTrue(discord_embed.footer)

    def test_should_set_special_footer_for_generated_notifications(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        notification = Notification.create_from_structure(
            structure, notif_type=NotificationType.STRUCTURE_FUEL_ALERT
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        self.assertEqual(discord_embed.footer.text, "Structures")
        self.assertIn("structures_logo.png", discord_embed.footer.icon_url)


class TestNotificationEmbedsClasses(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        user, _ = create_user_from_evecharacter(
            1001, permissions=["structures.add_structure_owner"]
        )
        cls.owner = create_owner_from_user(user=user)
        cls.moon_id = 40161465
        cls.solar_system_id = 30002537
        cls.type_id = 16213

    def test_should_generate_embed_for_normal_tower_resource_alert(self):
        # given
        create_starbase(
            owner=self.owner,
            eve_moon_id=self.moon_id,
            eve_solar_system_id=self.solar_system_id,
            eve_type_id=self.type_id,
        )
        data = {
            "corpID": self.owner.corporation.corporation_id,
            "moonID": self.moon_id,
            "solarSystemID": self.solar_system_id,
            "typeID": self.type_id,
            "wants": [{"quantity": 120, "typeID": 4051}],
        }
        notification = create_notification(
            owner=self.owner,
            notif_type=NotificationType.TOWER_RESOURCE_ALERT_MSG,
            data=data,
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        description = markdown_to_plain(discord_embed.description)
        self.assertIn("is running out of fuel in 3 hours", description)

    def test_should_generate_embed_for_generated_tower_resource_alert(self):
        # given
        structure = create_starbase(
            owner=self.owner,
            eve_moon_id=self.moon_id,
            eve_solar_system_id=self.solar_system_id,
            eve_type_id=self.type_id,
            fuel_expires_at=now() + dt.timedelta(hours=2, seconds=20),
        )
        notification = Notification.create_from_structure(
            structure=structure, notif_type=NotificationType.TOWER_RESOURCE_ALERT_MSG
        )
        notification_embed = NotificationBaseEmbed.create(notification)
        # when
        discord_embed = notification_embed.generate_embed()
        # then
        description = markdown_to_plain(discord_embed.description)
        self.assertIn("is running out of fuel in 2 hours", description)


class TestGeneratedNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_create_tower_reinforced_embed(self):
        # given
        notif = GeneratedNotificationFactory()
        # when
        obj = NotificationBaseEmbed.create(notif)
        # then
        self.assertIsInstance(obj, NotificationTowerReinforcedExtra)

    def test_should_generate_embed(self):
        # given
        notif = GeneratedNotificationFactory()
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertIsInstance(obj, dhooks_lite.Embed)
        starbase = notif.structures.first()
        self.assertIn(starbase.name, obj.description)


class TestEveNotificationEmbeds(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_create_sov_embed(self):
        # given
        notif = NotificationFactory(
            owner=self.owner,
            sender=EveEntityAllianceFactory(),
            notif_type=NotificationType.SOV_ENTOSIS_CAPTURE_STARTED,
            text_from_dict={"solarSystemID": 30000474, "structureTypeID": 32226},
        )
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertIsInstance(obj, dhooks_lite.Embed)

    def test_should_create_sov_embed_without_sender(self):
        # given
        notif = NotificationFactory(
            owner=self.owner,
            sender=None,
            notif_type=NotificationType.SOV_ENTOSIS_CAPTURE_STARTED,
            text_from_dict={"solarSystemID": 30000474, "structureTypeID": 32226},
        )
        embed = NotificationBaseEmbed.create(notif)
        # when
        obj = embed.generate_embed()
        # then
        self.assertIsInstance(obj, dhooks_lite.Embed)



"""functions for loading test data and for building mocks"""
import datetime as dt
import json
import math
import unicodedata
from copy import deepcopy
from pathlib import Path
from random import randrange
from typing import Optional, Tuple
from unittest.mock import Mock

from bravado.exception import HTTPNotFound
from bs4 import BeautifulSoup
from markdown import markdown

from django.contrib.auth.models import User
from django.utils.dateparse import parse_datetime
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from app_utils.esi_testing import BravadoOperationStub, BravadoResponseStub
from app_utils.testing import create_user_from_evecharacter

from structures.core.notification_types import NotificationType
from structures.models import (
    EveSovereigntyMap,
    Notification,
    Owner,
    Structure,
    StructureService,
    StructureTag,
    Webhook,
)

from .factories import create_notification

ESI_CORP_STRUCTURES_PAGE_SIZE = 2

_current_folder = Path(__file__).parent
_FILENAME_EVEUNIVERSE_TESTDATA = "eveuniverse.json"


def test_data_filename():
    return f"{_current_folder}/{_FILENAME_EVEUNIVERSE_TESTDATA}"


##############################
# internal functions


def _load_esi_data():
    with (_current_folder / "esi_data.json").open("r") as f:
        data = json.load(f)

    return data


def _load_testdata_entities() -> dict:
    with (_current_folder / "entities.json").open("r") as f:
        entities = json.load(f)

    # update timestamp to current
    for notification in entities["Notification"]:
        notification["timestamp"] = now() - dt.timedelta(
            hours=randrange(3), minutes=randrange(60), seconds=randrange(60)
        )

    # update timestamps on structures
    for structure in entities["Structure"]:
        if "fuel_expires_at" in structure:
            fuel_expires_at = now() + dt.timedelta(days=1 + randrange(5))
            structure["fuel_expires_at"] = fuel_expires_at

        if "state_timer_start" in structure:
            state_timer_start = now() + dt.timedelta(days=1 + randrange(3))
            structure["state_timer_start"] = state_timer_start
            state_timer_end = state_timer_start + dt.timedelta(minutes=15)
            structure["state_timer_end"] = state_timer_end

        if "reinforced_until" in structure:
            structure["reinforced_until"] = parse_datetime(
                structure["reinforced_until"]
            )

        if "unanchors_at" in structure:
            unanchors_at = now() + dt.timedelta(days=3 + randrange(5))
            structure["unanchors_at"] = unanchors_at

    return entities


esi_data = _load_esi_data()
esi_corp_structures_data = esi_data["Corporation"][
    "get_corporations_corporation_id_structures"
]
entities_testdata = _load_testdata_entities()


##############################
# functions for mocking calls to ESI with test data

ESI_LANGUAGES = {
    "de",
    "en-us",
    "fr",
    "ja",
    "ru",
    "ko",
    # "zh"
}


def esi_get_universe_planets_planet_id(planet_id, language=None, **kwargs):
    """simulates ESI endpoint of same name for mock test
    will use the respective test data
    unless the function property override_data is set
    """

    entity = None
    for x in entities_testdata["EvePlanet"]:
        if x["id"] == planet_id:
            entity = x.copy()
            break

    if entity is None:
        raise ValueError("entity with id {} not found in testdata".format(planet_id))

    entity["planet_id"] = entity.pop("id")
    entity["system_id"] = entity.pop("eve_solar_system_id")
    entity["type_id"] = entity.pop("eve_type_id")
    entity["position"] = {"x": 1, "y": 2, "z": 3}

    if language in ESI_LANGUAGES.difference({"en-us"}):
        entity["name"] += "_" + language

    return BravadoOperationStub(data=entity)


def esi_get_corporations_corporation_id_structures(
    corporation_id, token, page=None, language=None, **kwargs
):
    """simulates ESI endpoint of same name for mock test
    will use the respective test data
    unless the function property override_data is set
    """
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    page_size = ESI_CORP_STRUCTURES_PAGE_SIZE
    if not page:
        page = 1

    if esi_get_corporations_corporation_id_structures.override_data is None:
        my_corp_structures_data = esi_corp_structures_data
    else:
        if not isinstance(
            esi_get_corporations_corporation_id_structures.override_data, dict
        ):
            raise TypeError("data must be dict")

        my_corp_structures_data = (
            esi_get_corporations_corporation_id_structures.override_data
        )

    if str(corporation_id) in my_corp_structures_data:
        corp_data = deepcopy(my_corp_structures_data[str(corporation_id)])
    else:
        corp_data = list()

    # add pseudo localization
    if language:
        for obj in corp_data:
            if "services" in obj and obj["services"]:
                for service in obj["services"]:
                    if language != "en-us":
                        service["name"] += "_%s" % language

    # convert datetime
    for obj in corp_data:
        for key in obj:
            if isinstance(obj[key], str):
                my_dt = parse_datetime(obj[key])
                if my_dt:
                    obj[key] = my_dt

    start = (page - 1) * page_size
    stop = start + page_size
    pages_count = int(math.ceil(len(corp_data) / page_size))

    return BravadoOperationStub(
        data=corp_data[start:stop], headers={"x-pages": pages_count}
    )


esi_get_corporations_corporation_id_structures.override_data = None


def esi_get_corporations_corporation_id_structures_2(
    corporation_id, token, page=None, language=None, **kwargs
):
    """simulates ESI endpoint of same name for mock test
    will use the respective test data
    unless the function property override_data is set

    VARIANT that simulates django-esi 2.0
    """

    if not isinstance(token, str):
        raise ValueError("token must be a string")

    page_size = ESI_CORP_STRUCTURES_PAGE_SIZE
    if not page:
        page = 1

    if esi_get_corporations_corporation_id_structures.override_data is None:
        my_corp_structures_data = esi_corp_structures_data
    else:
        if not isinstance(
            esi_get_corporations_corporation_id_structures.override_data, dict
        ):
            raise TypeError("data must be dict")

        my_corp_structures_data = (
            esi_get_corporations_corporation_id_structures.override_data
        )

    if str(corporation_id) in my_corp_structures_data:
        corp_data = deepcopy(my_corp_structures_data[str(corporation_id)])
    else:
        corp_data = list()

    # add pseudo localization
    if language:
        for obj in corp_data:
            if "services" in obj and obj["services"]:
                for service in obj["services"]:
                    if language != "en-us":
                        service["name"] += "_%s" % language

    # convert datetime
    for obj in corp_data:
        if "fuel_expires" in obj and obj["fuel_expires"]:
            obj["fuel_expires"] = parse_datetime(obj["fuel_expires"])

    start = (page - 1) * page_size
    stop = start + page_size
    pages_count = int(math.ceil(len(corp_data) / page_size))

    return BravadoOperationStub(
        data=corp_data[start:stop], headers={"x-pages": pages_count}
    )


def esi_get_corporations_corporation_id_starbases(
    corporation_id, token, page=None, **kwargs
):
    """simulates ESI endpoint of same name for mock test
    will use the respective test data
    unless the function property override_data is set
    """
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    page_size = ESI_CORP_STRUCTURES_PAGE_SIZE
    if not page:
        page = 1

    if esi_get_corporations_corporation_id_starbases.override_data is None:
        my_corp_starbases_data = esi_data["Corporation"][
            "get_corporations_corporation_id_starbases"
        ]
    else:
        if not isinstance(
            esi_get_corporations_corporation_id_starbases.override_data, dict
        ):
            raise TypeError("data must be dict")

        my_corp_starbases_data = (
            esi_get_corporations_corporation_id_starbases.override_data
        )

    if str(corporation_id) in my_corp_starbases_data:
        corp_data = deepcopy(my_corp_starbases_data[str(corporation_id)])
    else:
        corp_data = list()

    start = (page - 1) * page_size
    stop = start + page_size
    pages_count = int(math.ceil(len(corp_data) / page_size))
    return BravadoOperationStub(
        data=corp_data[start:stop], headers={"x-pages": pages_count}
    )


esi_get_corporations_corporation_id_starbases.override_data = None


def esi_get_corporations_corporation_id_starbases_starbase_id(
    corporation_id, starbase_id, system_id, token, *args, **kwargs
):
    """simulates ESI endpoint of same name for mock test"""
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    corporation_starbase_details = esi_data["Corporation"][
        "get_corporations_corporation_id_starbases_starbase_id"
    ]  # noqa
    if str(starbase_id) not in corporation_starbase_details:
        mock_response = Mock()
        mock_response.status_code = 404
        message = "Can not find starbase with ID %s" % starbase_id
        raise HTTPNotFound(mock_response, message=message)

    return BravadoOperationStub(
        data=corporation_starbase_details[str(starbase_id)],
        headers={
            "Last-Modified": dt.datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")
        },
    )


def esi_get_universe_structures_structure_id(structure_id, token, *args, **kwargs):
    """simulates ESI endpoint of same name for mock test"""
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    if not esi_get_universe_structures_structure_id.override_data:
        universe_structures_data = esi_data["Universe"][
            "get_universe_structures_structure_id"
        ]
    else:
        universe_structures_data = (
            esi_get_universe_structures_structure_id.override_data
        )

    if str(structure_id) in universe_structures_data:
        return BravadoOperationStub(data=universe_structures_data[str(structure_id)])

    else:
        mock_response = Mock()
        mock_response.status_code = 404
        message = "Can not find structure with ID %s" % structure_id
        raise HTTPNotFound(mock_response, message=message)


esi_get_universe_structures_structure_id.override_data = None


def esi_get_characters_character_id_notifications(character_id, token, *args, **kwargs):
    """simulates ESI endpoint of same name for mock test"""
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    return BravadoOperationStub(data=entities_testdata["Notification"])


def esi_get_corporations_corporation_id_customs_offices(
    corporation_id, token, page=None, **kwargs
):
    """simulates ESI endpoint of same name for mock test
    will use the respective test data
    unless the function property override_data is set
    """
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    page_size = ESI_CORP_STRUCTURES_PAGE_SIZE
    if not page:
        page = 1

    if (
        esi_get_corporations_corporation_id_customs_offices.override_data is None
    ):  # noqa: E501
        my_corp_customs_offices_data = esi_data["Planetary_Interaction"][
            "get_corporations_corporation_id_customs_offices"
        ]

    else:
        if not isinstance(
            esi_get_corporations_corporation_id_customs_offices.override_data, dict
        ):
            raise TypeError("data must be dict")

        my_corp_customs_offices_data = (
            esi_get_corporations_corporation_id_customs_offices.override_data
        )

    if str(corporation_id) in my_corp_customs_offices_data:
        corp_data = deepcopy(my_corp_customs_offices_data[str(corporation_id)])
    else:
        corp_data = list()

    start = (page - 1) * page_size
    stop = start + page_size
    pages_count = int(math.ceil(len(corp_data) / page_size))
    return BravadoOperationStub(
        data=corp_data[start:stop], headers={"x-pages": pages_count}
    )


esi_get_corporations_corporation_id_customs_offices.override_data = None


def esi_get_corporations_corporation_id_assets(
    corporation_id: int, token: str, **kwargs
):
    """simulates ESI endpoint of same name for mock test"""
    try:
        my_esi_data = esi_data["Assets"]["get_corporations_corporation_id_assets"][
            str(corporation_id)
        ]
    except KeyError:
        raise HTTPNotFound(
            response=BravadoResponseStub(
                404, f"No asset data found for {corporation_id} "
            )
        ) from None
    return BravadoOperationStub(data=my_esi_data)


def _esi_post_corporations_corporation_id_assets(
    category: str,
    corporation_id: int,
    item_ids: list,
    my_esi_data: Optional[list] = None,
) -> list:
    """simulates ESI endpoint of same name for mock test"""

    if my_esi_data is None:
        my_esi_data = esi_data["Corporation"][category]

    if str(corporation_id) not in my_esi_data:
        raise RuntimeError(
            "No asset data found for corporation {} in {}".format(
                corporation_id, category
            )
        )
    else:
        return BravadoOperationStub(data=my_esi_data[str(corporation_id)])


def esi_post_corporations_corporation_id_assets_locations(
    corporation_id: int, item_ids: list, token, *args, **kwargs
) -> list:
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    return _esi_post_corporations_corporation_id_assets(
        "post_corporations_corporation_id_assets_locations",
        corporation_id,
        item_ids,
        esi_post_corporations_corporation_id_assets_locations.override_data,
    )


esi_post_corporations_corporation_id_assets_locations.override_data = None


def esi_post_corporations_corporation_id_assets_names(
    corporation_id: int, item_ids: list, token: str, *args, **kwargs
) -> list:
    if not isinstance(token, str):
        raise ValueError("token must be a string")
    return _esi_post_corporations_corporation_id_assets(
        "post_corporations_corporation_id_assets_names",
        corporation_id,
        item_ids,
        esi_post_corporations_corporation_id_assets_names.override_data,
    )


esi_post_corporations_corporation_id_assets_names.override_data = None


def esi_get_universe_categories_category_id(category_id, language=None):
    obj_data = {"id": 65, "name": "Structure"}
    if language in ESI_LANGUAGES.difference({"en-us"}):
        obj_data["name"] += "_" + language

    return BravadoOperationStub(data=obj_data)


def esi_get_universe_moons_moon_id(moon_id, language=None):
    obj_data = {
        "id": 40161465,
        "name": "Amamake II - Moon 1",
        "system_id": 30002537,
        "position": {"x": 1, "y": 2, "z": 3},
    }
    if language in ESI_LANGUAGES.difference({"en-us"}):
        obj_data["name"] += "_" + language

    return BravadoOperationStub(data=obj_data)


def esi_return_data(data):
    return lambda **kwargs: BravadoOperationStub(data=data)


def esi_mock_client(version=1.6):
    """provides a mocked ESI client

    version is the supported version of django-esi
    """
    mock_client = Mock()

    # Assets
    mock_client.Assets.get_corporations_corporation_id_assets = (
        esi_get_corporations_corporation_id_assets
    )
    mock_client.Assets.post_corporations_corporation_id_assets_locations = (
        esi_post_corporations_corporation_id_assets_locations
    )
    mock_client.Assets.post_corporations_corporation_id_assets_names = (
        esi_post_corporations_corporation_id_assets_names
    )
    # Character
    mock_client.Character.get_characters_character_id_notifications.side_effect = (
        esi_get_characters_character_id_notifications
    )
    # Corporation
    mock_client.Corporation.get_corporations_corporation_id_structures.side_effect = (
        esi_get_corporations_corporation_id_structures_2
    )
    # mock_client.Corporation.get_corporations_corporation_id_starbases.side_effect = (
    #     RuntimeError
    # )
    # mock_client.Corporation.get_corporations_corporation_id_starbases_starbase_id.side_effect = (
    #     esi_get_corporations_corporation_id_starbases_starbase_id
    # )
    # Planetary Interaction
    mock_client.Planetary_Interaction.get_corporations_corporation_id_customs_offices = (
        esi_get_corporations_corporation_id_customs_offices
    )
    # Sovereignty
    mock_client.Sovereignty.get_sovereignty_map.side_effect = esi_return_data(
        esi_data["Sovereignty"]["get_sovereignty_map"]
    )
    # Universe
    mock_client.Universe.get_universe_categories_category_id.side_effect = (
        esi_get_universe_categories_category_id
    )

    mock_client.Universe.get_universe_groups_group_id.side_effect = esi_return_data(
        {
            "id": 1657,
            "name": "Citadel",
            "category_id": 65,
        }
    )
    mock_client.Universe.get_universe_types_type_id.side_effect = esi_return_data(
        {
            "id": 35832,
            "name": "Astrahus",
            "group_id": 1657,
        }
    )
    mock_client.Universe.get_universe_regions_region_id.side_effect = esi_return_data(
        {
            "id": 10000005,
            "name": "Detorid",
        }
    )
    mock_client.Universe.get_universe_constellations_constellation_id.side_effect = (
        esi_return_data(
            {
                "id": 20000069,
                "name": "1RG-GU",
                "region_id": 10000005,
            }
        )
    )
    mock_client.Universe.get_universe_systems.side_effect = esi_return_data(
        esi_data["Universe"]["get_universe_systems"]
    )
    mock_client.Universe.get_universe_systems_system_id.side_effect = esi_return_data(
        {
            "id": 30000474,
            "name": "1-PGSG",
            "security_status": -0.496552765369415,
            "constellation_id": 20000069,
            "star_id": 99,
            "planets": [
                {"planet_id": 40029526},
                {"planet_id": 40029528},
                {"planet_id": 40029529},
            ],
        }
    )
    mock_client.Universe.get_universe_planets_planet_id.side_effect = (
        esi_get_universe_planets_planet_id
    )
    mock_client.Universe.get_universe_moons_moon_id.side_effect = (
        esi_get_universe_moons_moon_id
    )
    mock_client.Universe.post_universe_names.side_effect = esi_return_data(
        [{"id": 3011, "category": "alliance", "name": "Big Bad Alliance"}]
    )
    mock_client.Universe.get_universe_structures_structure_id.side_effect = (
        esi_get_universe_structures_structure_id
    )
    return mock_client


###################################
# helper functions
#


def load_entity(EntityClass):
    """loads testdata for given entity class"""
    entity_name = EntityClass.__name__
    EntityClass.objects.all().delete()
    for obj in entities_testdata[entity_name]:
        if EntityClass is EveCharacter:
            EveCharacter.objects.create(**obj)
            corp_defaults = {
                "corporation_name": obj["corporation_name"],
                "member_count": 99,
            }
            if "alliance_id" in obj:
                alliance, _ = EveAllianceInfo.objects.get_or_create(
                    alliance_id=obj["alliance_id"],
                    defaults={
                        "alliance_name": obj["alliance_name"],
                        "executor_corp_id": obj["corporation_id"],
                    },
                )
                corp_defaults["alliance"] = alliance
            EveCorporationInfo.objects.get_or_create(
                corporation_id=obj["corporation_id"], defaults=corp_defaults
            )
            continue
        elif EntityClass is Webhook:
            obj["notification_types"] = NotificationType.values
        EntityClass.objects.create(**obj)
    assert len(entities_testdata[entity_name]) == EntityClass.objects.count()


def load_entities(entities_def: Optional[list] = None):
    """loads testdata for given entities classes"""
    entities_def_master = [
        EveSovereigntyMap,
        EveCharacter,
        EveEntity,
        StructureTag,
        Webhook,
    ]
    for EntityClass in entities_def_master:
        if not entities_def or EntityClass in entities_def:
            load_entity(EntityClass)


def create_structures(dont_load_entities: bool = False) -> object:
    """create structure entities from test data"""
    if not dont_load_entities:
        load_entities()

    create_owners()

    for character in EveCharacter.objects.all():
        EveEntity.objects.get_or_create(
            id=character.character_id,
            defaults={
                "category": EveEntity.CATEGORY_CHARACTER,
                "name": character.character_name,
            },
        )
        corporation = EveCorporationInfo.objects.get(
            corporation_id=character.corporation_id
        )
        if corporation.alliance:
            character.alliance_id = corporation.alliance.alliance_id
            character.alliance_name = corporation.alliance.alliance_name
            character.save()

    StructureTag.objects.get(name="tag_a")
    tag_b = StructureTag.objects.get(name="tag_b")
    tag_c = StructureTag.objects.get(name="tag_c")
    Structure.objects.all().delete()
    for structure in entities_testdata["Structure"]:
        x = structure.copy()
        x["last_updated_at"] = now()
        x["owner"] = Owner.objects.get(
            corporation__corporation_id=x["owner_corporation_id"]
        )
        del x["owner_corporation_id"]

        if "services" in x:
            del x["services"]

        obj = Structure.objects.create(**x)
        if obj.state != 11:
            obj.state_timer_start = now() - dt.timedelta(days=randrange(3) + 1)
            obj.state_timer_start = obj.state_timer_start + dt.timedelta(
                days=randrange(4) + 1
            )

        if obj.id in [1000000000002, 1000000000003]:
            obj.tags.add(tag_c)

        if obj.id in [1000000000003]:
            obj.tags.add(tag_b)

        if "services" in structure:
            for service in structure["services"]:
                StructureService.objects.create(
                    structure=obj,
                    name=service["name"],
                    state=StructureService.State.from_esi_name(service["state"]),
                )
        obj.save()


def create_owners():
    default_webhooks = Webhook.objects.filter(is_default=True)
    for corporation in EveCorporationInfo.objects.all():
        EveEntity.objects.get_or_create(
            id=corporation.corporation_id,
            defaults={
                "category": EveEntity.CATEGORY_CORPORATION,
                "name": corporation.corporation_name,
            },
        )
        my_owner = Owner.objects.create(corporation=corporation)
        for x in default_webhooks:
            my_owner.webhooks.add(x)

        if int(corporation.corporation_id) in [2001, 2002]:
            alliance = EveAllianceInfo.objects.get(alliance_id=3001)
            corporation.alliance = alliance
            corporation.save()


def create_user(character_id, load_data=False) -> User:
    """create a user from the given character id and returns it

    Needs: EveCharacter
    """
    if load_data:
        load_entity(EveCharacter)

    my_character = EveCharacter.objects.get(character_id=character_id)
    my_user = User.objects.create_user(
        my_character.character_name, "abc@example.com", "password"
    )
    CharacterOwnership.objects.create(
        character=my_character,
        owner_hash="x1" + my_character.character_name,
        user=my_user,
    )
    my_user.profile.main_character = my_character
    return my_user


def set_owner_character(character_id: int) -> Tuple[User, Owner]:
    """Set owner character for the owner related to the given character.
    Creates a new user.
    """
    my_user, character_ownership = create_user_from_evecharacter(
        character_id,
        permissions=["structures.add_structure_owner"],
        scopes=Owner.get_esi_scopes(),
    )
    my_character = my_user.profile.main_character
    my_owner = Owner.objects.get(
        corporation__corporation_id=my_character.corporation_id
    )
    my_owner.characters.create(character_ownership=character_ownership)
    return my_user, my_owner


def load_notification_by_type(
    owner: Owner, notif_type: NotificationType
) -> Notification:
    for notification in entities_testdata["Notification"]:
        if notification["type"] == notif_type.value:
            return create_notification(
                owner=owner,
                notif_type=notif_type.value,
                text=notification.get("text", ""),
            )
    raise ValueError(f"Could not find notif for type: {notif_type}")


def load_notification_entities(owner: Owner):
    timestamp_start = now() - dt.timedelta(hours=2)
    for notification in entities_testdata["Notification"]:
        _load_notification_for_owner(owner, timestamp_start, notification)


def _load_notification_for_owner(owner, timestamp_start, notification, new_ids=False):
    notification_id = notification["notification_id"]
    notif_type = notification["type"]
    # print(f"Reading notification: {notification_id}-{notif_type}")
    sender = EveEntity.objects.get(id=notification["sender_id"])
    text = notification["text"] if "text" in notification else None
    is_read = notification["is_read"] if "is_read" in notification else None
    timestamp_start = timestamp_start + dt.timedelta(minutes=5)
    Notification.objects.update_or_create(
        notification_id=notification_id,
        owner=owner,
        defaults={
            "sender": sender,
            "timestamp": timestamp_start,
            "notif_type": notif_type,
            "text": text,
            "is_read": is_read,
            "last_updated": now(),
            "is_sent": False,
        },
    )


def markdown_to_plain(text: str) -> str:
    """Convert text in markdown to plain text."""
    html = markdown(text)
    text = "".join(BeautifulSoup(html, features="html.parser").findAll(text=True))
    return unicodedata.normalize("NFKD", text)



import datetime as dt
from copy import deepcopy
from unittest.mock import patch

from pytz import UTC

from django.utils.timezone import now
from eveuniverse.models import EveSolarSystem

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from structures.constants import EveTypeId
from structures.core.notification_types import NotificationType
from structures.models import (
    EveSovereigntyMap,
    FuelAlertConfig,
    JumpFuelAlertConfig,
    PocoDetails,
    Structure,
    StructureItem,
    StructureService,
    StructureTag,
)
from structures.tests.testdata.factories import (
    create_owner_from_user,
    create_starbase,
    create_upwell_structure,
)
from structures.tests.testdata.factories_2 import (
    JumpGateFactory,
    OwnerFactory,
    PocoFactory,
    StarbaseFactory,
    StructureFactory,
    StructureTagFactory,
)
from structures.tests.testdata.helpers import (
    create_structures,
    load_entities,
    set_owner_character,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

STRUCTURES_PATH = "structures.models.structures_1"
NOTIFICATIONS_PATH = "structures.models.notifications"

EVE_ID_HELIUM_FUEL_BLOCK = 4247
EVE_ID_NITROGEN_FUEL_BLOCK = 4051


class TestStructureTag(NoSocketsTestCase):
    def test_str(self):
        obj = StructureTag(name="Super cool tag")
        self.assertEqual(str(obj), "Super cool tag")

    def test_repr(self):
        obj = StructureTag.objects.create(name="Super cool tag")
        expected = "StructureTag(name='Super cool tag')"
        self.assertEqual(repr(obj), expected)

    def test_list_sorted(self):
        x1 = StructureTag(name="Alpha")
        x2 = StructureTag(name="charlie")
        x3 = StructureTag(name="bravo")
        tags = [x1, x2, x3]

        self.assertListEqual(StructureTag.sorted(tags), [x1, x3, x2])
        self.assertListEqual(StructureTag.sorted(tags, reverse=True), [x2, x3, x1])

    def test_html_default(self):
        x = StructureTag(name="Super cool tag")
        self.assertEqual(
            x.html, '<span class="label label-default">Super cool tag</span>'
        )

    def test_html_primary(self):
        x = StructureTag(name="Super cool tag", style="primary")
        self.assertEqual(
            x.html, '<span class="label label-primary">Super cool tag</span>'
        )


class TestStructure(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        _, cls.owner = set_owner_character(character_id=1001)

    def test_str_upwell(self):
        obj = Structure.objects.get(id=1000000000001)
        expected = "Amamake - Test Structure Alpha"
        self.assertEqual(str(obj), expected)

    def test_str_poco(self):
        obj = Structure.objects.get(id=1200000000003)
        expected = "Amamake V - Customs Office (Amamake V)"
        self.assertEqual(str(obj), expected)

    def test_str_starbase(self):
        obj = Structure.objects.get(id=1300000000001)
        expected = "Amamake II - Moon 1 - Home Sweat Home"
        self.assertEqual(str(obj), expected)

    def test_repr_upwell(self):
        obj = Structure.objects.get(id=1000000000001)
        expected = (
            "Structure(id=1000000000001, "
            "eve_solar_system='Amamake', "
            "eve_planet='', "
            "eve_moon='', "
            "name='Test Structure Alpha')"
        )
        self.assertEqual(repr(obj), expected)

    def test_repr_poco(self):
        obj = Structure.objects.get(id=1200000000003)
        expected = (
            "Structure(id=1200000000003, "
            "eve_solar_system='Amamake', "
            "eve_planet='Amamake V', "
            "eve_moon='', "
            "name='Customs Office (Amamake V)')"
        )
        self.assertEqual(repr(obj), expected)

    def test_repr_starbase(self):
        obj = Structure.objects.get(id=1300000000001)
        expected = (
            "Structure(id=1300000000001, "
            "eve_solar_system='Amamake', "
            "eve_planet='', "
            "eve_moon='Amamake II - Moon 1', "
            "name='Home Sweat Home')"
        )
        self.assertEqual(repr(obj), expected)

    def test_is_full_power(self):
        structure = Structure.objects.get(id=1000000000001)
        poco = Structure.objects.get(id=1200000000003)

        # true when upwell structure and has fuel that is not expired
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        self.assertTrue(structure.is_full_power)

        # false when upwell structure and has fuel, but is expired
        structure.fuel_expires_at = now() - dt.timedelta(hours=1)
        self.assertFalse(structure.is_full_power)

        # False when no fuel info
        structure.fuel_expires_at = None
        self.assertFalse(structure.is_full_power)

        # none when no upwell structure
        poco.fuel_expires_at = now() + dt.timedelta(hours=1)
        self.assertIsNone(poco.is_full_power)

    def test_is_low_power(self):
        structure = Structure.objects.get(id=1000000000001)

        # true if Upwell structure and fuel expired and last online < 7d
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertTrue(structure.is_low_power)

        # True if Upwell structure and no fuel info and last online < 7d
        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertTrue(structure.is_low_power)

        # false if Upwell structure and it has fuel
        structure.fuel_expires_at = now() + dt.timedelta(days=3)
        self.assertFalse(structure.is_low_power)

        # none if upwell structure, but not online info
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = None
        self.assertFalse(structure.is_low_power)

        structure.fuel_expires_at = None
        structure.last_online_at = None
        self.assertFalse(structure.is_low_power)

        # none for non structures
        starbase = Structure.objects.get(id=1300000000001)
        self.assertIsNone(starbase.is_low_power)

        pos = Structure.objects.get(id=1200000000003)
        self.assertIsNone(pos.is_low_power)

    def test_is_abandoned(self):
        # none for non structures
        starbase = Structure.objects.get(id=1300000000001)  # starbase
        self.assertIsNone(starbase.is_abandoned)

        structure = Structure.objects.get(id=1000000000001)

        # true when upwell structure, online > 7 days
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=1)

        # false when upwell structure, online <= 7 days or none
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=0)
        self.assertFalse(structure.is_abandoned)

        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertFalse(structure.is_abandoned)

        # none if missing information
        structure.last_online_at = None
        self.assertFalse(structure.is_abandoned)

    def test_structure_service_str(self):
        structure = Structure.objects.get(id=1000000000001)
        obj = StructureService(
            structure=structure, name="Dummy", state=StructureService.State.ONLINE
        )
        self.assertEqual(str(obj), "Amamake - Test Structure Alpha - Dummy")

    def test_extract_name_from_esi_response(self):
        expected = "Alpha"
        self.assertEqual(
            Structure.extract_name_from_esi_response("Super - Alpha"), expected
        )
        self.assertEqual(Structure.extract_name_from_esi_response("Alpha"), expected)

    def test_owner_has_sov(self):
        # Wayne Tech has sov in 1-PG
        pos = Structure.objects.get(id=1300000000003)
        self.assertTrue(pos.owner_has_sov)

        # Wayne Tech has no sov in A-C5TC
        structure = Structure.objects.get(id=1000000000003)
        self.assertFalse(structure.owner_has_sov)

        # Wayne Tech has no sov in Amamake
        structure = Structure.objects.get(id=1000000000001)
        self.assertFalse(structure.owner_has_sov)

    def test_should_return_hours_when_fuel_expires(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=2)
        # when
        result = structure.hours_fuel_expires
        # then
        self.assertAlmostEqual(result, 2.0, delta=0.1)

    def test_should_return_none_when_not_fuel_info(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = None
        # when
        result = structure.hours_fuel_expires
        # then
        self.assertIsNone(result)

    def test_should_return_moon_location(self):
        # given
        starbase = Structure.objects.get(id=1300000000001)
        # when/then
        self.assertEqual(starbase.location_name, "Amamake II - Moon 1")

    def test_should_return_planet_location(self):
        # given
        poco = Structure.objects.get(id=1200000000003)
        # when/then
        self.assertEqual(poco.location_name, "Amamake V")

    def test_should_return_solar_system_location(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        # when/then
        self.assertEqual(structure.location_name, "Amamake")

    def test_is_poco(self):
        # given
        structure = StructureFactory.build()
        poco = PocoFactory.build()
        starbase = StarbaseFactory.build()
        # then
        self.assertFalse(structure.is_poco)
        self.assertTrue(poco.is_poco)
        self.assertFalse(starbase.is_poco)

    def test_is_starbase(self):
        # given
        structure = StructureFactory.build()
        poco = PocoFactory.build()
        starbase = StarbaseFactory.build()
        # then
        self.assertFalse(structure.is_starbase)
        self.assertFalse(poco.is_starbase)
        self.assertTrue(starbase.is_starbase)

    def test_is_upwell_structure(self):
        # given
        structure = StructureFactory.build()
        poco = PocoFactory.build()
        starbase = StarbaseFactory.build()
        # then
        self.assertTrue(structure.is_upwell_structure)
        self.assertFalse(poco.is_upwell_structure)
        self.assertFalse(starbase.is_upwell_structure)

    def test_is_jump_gate(self):
        # given
        normal_structure = Structure.objects.get(id=1000000000001)
        poco = PocoFactory.build()
        starbase = StarbaseFactory.build()
        jump_gate = JumpGateFactory.build()
        # then
        self.assertFalse(normal_structure.is_jump_gate)
        self.assertFalse(poco.is_jump_gate)
        self.assertFalse(starbase.is_jump_gate)
        self.assertTrue(jump_gate.is_jump_gate)

    # TODO: activate
    # def test_is_upwell_structure_data_error(self):
    #     # group without a category
    #     my_group = EveGroup.objects.create(id=299999, name="invalid group")
    #     my_type = EveType.objects.create(
    #         id=199999, name="invalid type", eve_group=my_group
    #     )
    #     self.assertFalse(my_type.is_upwell_structure)

    def test_should_detect_structure_has_position(self):
        # given
        structure = StructureFactory.build()
        # then
        self.assertTrue(structure.has_position)

    def test_should_detect_structure_has_no_position(self):
        # given
        structure = StructureFactory.build(
            position_x=None, position_y=None, position_z=None
        )
        # then
        self.assertFalse(structure.has_position)

    def test_should_return_jump_fuel_quantity(self):
        # given
        structure = Structure.objects.get(id=1000000000004)
        structure.items.create(
            id=1,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=32,
        )
        structure.items.create(
            id=2,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=10,
        )
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertEqual(result, 42)

    def test_should_return_none_for_jump_fuel_1(self):
        # given
        structure = Structure.objects.get(id=1000000000003)
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertIsNone(result)

    def test_should_return_none_for_jump_fuel_2(self):
        # given
        structure = Structure.objects.get(id=1200000000005)
        # when
        result = structure.jump_fuel_quantity()
        # then
        self.assertIsNone(result)

    def test_should_remove_fuel_alerts_when_fuel_level_above_threshold(self):
        # given
        config = JumpFuelAlertConfig.objects.create(threshold=100)
        structure = Structure.objects.get(id=1000000000004)
        structure.items.create(
            id=1,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=101,
        )
        structure.jump_fuel_alerts.create(config=config)
        # when
        structure.reevaluate_jump_fuel_alerts()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    def test_should_keep_fuel_alerts_when_fuel_level_below_threshold(self):
        # given
        config = JumpFuelAlertConfig.objects.create(threshold=100)
        structure = Structure.objects.get(id=1000000000004)
        structure.items.create(
            id=1,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=99,
        )
        structure.jump_fuel_alerts.create(config=config)
        # when
        structure.reevaluate_jump_fuel_alerts()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 1)

    def test_should_return_fuel_blocks(self):
        # given
        structure = Structure.objects.get(id=1000000000004)
        structure.items.create(
            id=1,
            eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=250,
        )
        structure.items.create(
            id=2,
            eve_type_id=EVE_ID_HELIUM_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=1000,
        )
        structure.items.create(
            id=3,
            eve_type_id=EVE_ID_HELIUM_FUEL_BLOCK,
            location_flag=StructureItem.LocationFlag.CARGO,
            is_singleton=False,
            quantity=500,
        )
        # when
        result = structure.structure_fuel_quantity
        # then
        self.assertEqual(result, 1250)

    def test_should_return_fuel_need(self):
        # given
        structure = Structure.objects.create(
            id=1900000000001,
            owner=self.owner,
            eve_solar_system_id=30002537,
            name="Dummy",
            state=Structure.State.SHIELD_VULNERABLE,
            eve_type_id=35832,
            fuel_expires_at=dt.datetime(2022, 1, 24, 5, 0, tzinfo=UTC),
        )
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 17, 15, 13, tzinfo=UTC)
            structure.items.create(
                id=1,
                eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
                location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                is_singleton=False,
                quantity=6309,
            )
        # when
        result = structure.structure_fuel_usage()
        # then
        self.assertEqual(result, 168)

    def test_should_not_break_when_no_fuel_date(self):
        # given
        structure = Structure.objects.create(
            id=1900000000001,
            owner=self.owner,
            eve_solar_system_id=30002537,
            name="Dummy",
            state=Structure.State.SHIELD_VULNERABLE,
            eve_type_id=35832,
            fuel_expires_at=None,
        )
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = dt.datetime(2021, 12, 17, 15, 13, tzinfo=UTC)
            structure.items.create(
                id=1,
                eve_type_id=EVE_ID_NITROGEN_FUEL_BLOCK,
                location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
                is_singleton=False,
                quantity=6309,
            )
        # when
        result = structure.structure_fuel_usage()
        # then
        self.assertIsNone(result)


class TestStructure3(NoSocketsTestCase):
    """New version of TestStructure using factories to create testdata."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.user, _ = create_user_from_evecharacter(1001)
        cls.owner = create_owner_from_user(cls.user)

    def test_should_not_show_structure_as_reinforced(self):
        # given
        structure = create_upwell_structure(
            owner=self.owner, state=Structure.State.SHIELD_VULNERABLE
        )
        # when/then
        self.assertFalse(structure.is_reinforced)

    def test_should_show_structure_as_reinforced(self):
        for state in [
            Structure.State.ARMOR_REINFORCE,
            Structure.State.HULL_REINFORCE,
            Structure.State.ANCHOR_VULNERABLE,
            Structure.State.HULL_VULNERABLE,
        ]:
            with self.subTest(state=state):
                structure = create_upwell_structure(owner=self.owner, state=state)
                self.assertTrue(structure.is_reinforced)

    def test_should_show_starbase_as_reinforced(self):
        # given
        structure = create_starbase(
            owner=self.owner, state=Structure.State.POS_REINFORCED
        )
        # when/then
        self.assertTrue(structure.is_reinforced)


class TestStructureIsBurningFuel(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        set_owner_character(character_id=1001)

    def test_should_return_true_for_structure(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        # when/then
        self.assertTrue(structure.is_burning_fuel)

    def test_should_return_false_for_structure(self):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = None
        # when/then
        self.assertFalse(structure.is_burning_fuel)

    def test_should_return_true_for_starbase(self):
        # given
        starbase = Structure.objects.get(id=1300000000001)
        for state in [
            Structure.State.POS_ONLINE,
            Structure.State.POS_REINFORCED,
            Structure.State.POS_UNANCHORING,
        ]:
            starbase.state = state
            # when/then
            self.assertTrue(starbase.is_burning_fuel)

    def test_should_return_false_for_starbase(self):
        # given
        starbase = Structure.objects.get(id=1300000000001)
        for state in [Structure.State.POS_OFFLINE, Structure.State.POS_ONLINING]:
            starbase.state = state
            # when/then
            self.assertFalse(starbase.is_burning_fuel)

    def test_should_return_false_for_poco(self):
        # given
        poco = Structure.objects.get(id=1200000000003)
        # when/then
        self.assertFalse(poco.is_burning_fuel)


@patch(STRUCTURES_PATH + ".STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True)
@patch(STRUCTURES_PATH + ".Structure.FUEL_DATES_EQUAL_THRESHOLD_UPWELL", 900)
@patch(STRUCTURES_PATH + ".Structure.FUEL_DATES_EQUAL_THRESHOLD_STARBASE", 7200)
class TestStructureFuelLevels(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        set_owner_character(character_id=1001)

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_reset_fuel_notifications_when_refueled_1(self):
        # given
        config = FuelAlertConfig.objects.create(start=48, end=0, repeat=12)
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.save()
        structure.structure_fuel_alerts.create(config=config, hours=12)
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=13)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_reset_fuel_notifications_when_fuel_expires_date_has_changed(self):
        # given
        config = FuelAlertConfig.objects.create(start=48, end=0, repeat=12)
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.save()
        old_instance = deepcopy(structure)
        structure.structure_fuel_alerts.create(config=config, hours=12)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=11)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(
        NOTIFICATIONS_PATH + ".Notification.send_to_configured_webhooks",
        lambda *args, **kwargs: None,
    )
    def test_should_not_reset_fuel_notifications_when_fuel_expiry_dates_unchanged(self):
        # given
        config = FuelAlertConfig.objects.create(start=48, end=0, repeat=12)
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.save()
        structure.structure_fuel_alerts.create(config=config, hours=12)
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=12, minutes=5)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 1)

    @patch(NOTIFICATIONS_PATH + ".Notification.create_from_structure")
    def test_should_generate_structure_refueled_notif_when_fuel_level_increased(
        self, mock_create_from_structure
    ):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=6)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_create_from_structure.called)
        _, kwargs = mock_create_from_structure.call_args
        self.assertEqual(
            kwargs["notif_type"], NotificationType.STRUCTURE_REFUELED_EXTRA
        )

    @patch(NOTIFICATIONS_PATH + ".Notification.create_from_structure")
    def test_should_generate_tower_refueled_notif_when_fuel_level_increased(
        self, mock_create_from_structure
    ):
        # given
        structure = Structure.objects.get(id=1300000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=4)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_create_from_structure.called)
        _, kwargs = mock_create_from_structure.call_args
        self.assertEqual(kwargs["notif_type"], NotificationType.TOWER_REFUELED_EXTRA)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_generate_refueled_notif_when_fuel_level_increased(
        self, mock_send_to_webhook
    ):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertTrue(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_when_fuel_level_almost_unchanged(
        self, mock_send_to_webhook
    ):
        # given
        structure = Structure.objects.get(id=1000000000001)
        target_date_1 = now() + dt.timedelta(hours=2)
        target_date_1 = now() + dt.timedelta(hours=2, minutes=15)
        structure.fuel_expires_at = target_date_1
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = target_date_1
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_fuel_level_decreased(
        self, mock_send_to_webhook
    ):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=12)
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)

    @patch(NOTIFICATIONS_PATH + ".Notification.send_to_webhook")
    def test_should_not_generate_refueled_notif_fuel_is_removed(
        self, mock_send_to_webhook
    ):
        # given
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=2)
        structure.save()
        old_instance = deepcopy(structure)
        # when
        structure.fuel_expires_at = None
        structure.handle_fuel_notifications(old_instance)
        # then
        self.assertFalse(mock_send_to_webhook.called)


class TestStructurePowerMode(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()

    def test_returns_none_for_non_upwell_structures(self):
        starbase = Structure.objects.get(id=1300000000001)
        self.assertIsNone(starbase.power_mode)

        pos = Structure.objects.get(id=1200000000003)
        self.assertIsNone(pos.power_mode)

        structure = Structure.objects.get(id=1000000000001)
        self.assertIsNotNone(structure.power_mode)

    def test_full_power_mode(self):
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        self.assertEqual(structure.power_mode, Structure.PowerMode.FULL_POWER)
        self.assertEqual(structure.get_power_mode_display(), "Full Power")

    def test_low_power_mode(self):
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)

        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = None
        structure.last_online_at = None
        structure.state = Structure.State.ANCHORING
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)

        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=3)
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_POWER)
        self.assertEqual(structure.get_power_mode_display(), "Low Power")

    def test_abandoned_mode(self):
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=1)
        self.assertEqual(structure.power_mode, Structure.PowerMode.ABANDONED)

        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=1)
        self.assertEqual(structure.power_mode, Structure.PowerMode.ABANDONED)
        self.assertEqual(structure.get_power_mode_display(), "Abandoned")

    def test_low_abandoned_mode(self):
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = now() - dt.timedelta(seconds=3)
        structure.last_online_at = None
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_ABANDONED)

        structure.fuel_expires_at = None
        structure.last_online_at = None
        self.assertEqual(structure.power_mode, Structure.PowerMode.LOW_ABANDONED)
        self.assertEqual(structure.get_power_mode_display(), "Abandoned?")


class TestStructure2(NoSocketsTestCase):
    def setUp(self):
        load_eveuniverse()
        load_entities([EveSovereigntyMap])
        create_structures()
        set_owner_character(character_id=1001)

    def test_can_create_generated_tags(self):
        obj = Structure.objects.get(id=1300000000003)
        obj.tags.clear()
        self.assertFalse(obj.tags.exists())
        obj.update_generated_tags()
        null_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(null_tag, list(obj.tags.all()))
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))

    def test_can_update_generated_tags(self):
        obj = Structure.objects.get(id=1300000000003)
        null_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(null_tag, list(obj.tags.all()))
        null_tag.order = 100
        null_tag.style = StructureTag.Style.DARK_BLUE
        null_tag.save()

        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, list(obj.tags.all()))
        sov_tag.order = 100
        sov_tag.style = StructureTag.Style.RED
        sov_tag.save()

        obj.update_generated_tags(recreate_tags=True)
        null_tag.refresh_from_db()
        self.assertEqual(null_tag.style, StructureTag.Style.RED)
        self.assertEqual(null_tag.order, 50)
        sov_tag.refresh_from_db()
        self.assertEqual(sov_tag.style, StructureTag.Style.DARK_BLUE)
        self.assertEqual(sov_tag.order, 20)


class TestStructureSave(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities([EveCharacter, EveSovereigntyMap])
        cls.owner = OwnerFactory(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )

    def test_can_save_tags_low_sec(self):
        obj = StructureFactory(
            owner=self.owner,
            eve_solar_system=EveSolarSystem.objects.get(name="Amamake"),
        )
        lowsec_tag = StructureTag.objects.get(name=StructureTag.NAME_LOWSEC_TAG)
        self.assertIn(lowsec_tag, obj.tags.all())
        self.assertIsNone(
            StructureTag.objects.filter(name=StructureTag.NAME_SOV_TAG).first()
        )

    def test_can_save_tags_null_sec_w_sov(self):
        obj = StructureFactory(
            owner=self.owner,
            eve_solar_system=EveSolarSystem.objects.get(name="1-PGSG"),
        )
        nullsec_tag = StructureTag.objects.get(name=StructureTag.NAME_NULLSEC_TAG)
        self.assertIn(nullsec_tag, obj.tags.all())
        sov_tag = StructureTag.objects.get(name=StructureTag.NAME_SOV_TAG)
        self.assertIn(sov_tag, obj.tags.all())

    def test_should_create_default_tags(self):
        # given
        default_tag = StructureTagFactory(is_default=True)
        # when
        obj = StructureFactory(owner=self.owner)
        # then
        self.assertIn(default_tag, obj.tags.all())

    def test_should_not_create_default_tags(self):
        # given
        default_tag = StructureTagFactory(is_default=True)
        # when
        obj = StructureFactory(owner=self.owner)
        obj.tags.all().delete()
        obj.save()
        # then
        self.assertNotIn(default_tag, obj.tags.all())


class TestStructureNoSetup(NoSocketsTestCase):
    def test_structure_get_matching_state(self):
        self.assertEqual(
            Structure.State.from_esi_name("anchoring"),
            Structure.State.ANCHORING,
        )
        self.assertEqual(
            Structure.State.from_esi_name("not matching name"),
            Structure.State.UNKNOWN,
        )

    def test_structure_service_get_matching_state(self):
        self.assertEqual(
            StructureService.State.from_esi_name("online"),
            StructureService.State.ONLINE,
        )
        self.assertEqual(
            StructureService.State.from_esi_name("offline"),
            StructureService.State.OFFLINE,
        )
        self.assertEqual(
            StructureService.State.from_esi_name("not matching"),
            StructureService.State.OFFLINE,
        )


class TestStructureService(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        set_owner_character(character_id=1001)

    def test_str(self):
        structure = Structure.objects.get(id=1000000000001)
        obj = StructureService.objects.get(structure=structure, name="Clone Bay")
        expected = "Amamake - Test Structure Alpha - Clone Bay"
        self.assertEqual(str(obj), expected)

    def test_repr(self):
        structure = Structure.objects.get(id=1000000000001)
        obj = StructureService.objects.get(structure=structure, name="Clone Bay")
        expected = "StructureService(structure_id=1000000000001, name='Clone Bay')"
        self.assertEqual(repr(obj), expected)


class TestPocoDetails(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        poco = cls.owner.structures.get(id=1200000000003)
        cls.details = PocoDetails.objects.create(
            structure=poco,
            alliance_tax_rate=0.02,
            allow_access_with_standings=True,
            allow_alliance_access=True,
            corporation_tax_rate=0.01,
            reinforce_exit_end=21,
            reinforce_exit_start=18,
        )

    def test_should_return_tax_for_corporation_member(self):
        # given
        my_character = EveCharacter.objects.get(character_id=1001)
        # when
        result = self.details.tax_for_character(my_character)
        # then
        self.assertEqual(result, 0.01)

    def test_should_return_tax_for_alliance_member(self):
        # given
        my_character = EveCharacter.objects.get(character_id=1003)
        # when
        result = self.details.tax_for_character(my_character)
        # then
        self.assertEqual(result, 0.02)

    def test_should_return_tax_for_unknown(self):
        # given
        my_character = EveCharacter.objects.get(character_id=1011)
        # when
        result = self.details.tax_for_character(my_character)
        # then
        self.assertIsNone(result)

    def test_should_return_standing_map_for_neutral_1(self):
        # given
        self.details.standing_level = PocoDetails.StandingLevel.NEUTRAL
        self.details.allow_access_with_standings = True
        # when
        result = self.details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": False,
                "BAD": False,
                "NEUTRAL": True,
                "GOOD": True,
                "EXCELLENT": True,
            },
        )

    def test_should_return_standing_map_for_neutral_2(self):
        # given
        self.details.standing_level = PocoDetails.StandingLevel.NEUTRAL
        self.details.allow_access_with_standings = False
        # when
        result = self.details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": False,
                "BAD": False,
                "NEUTRAL": False,
                "GOOD": False,
                "EXCELLENT": False,
            },
        )

    def test_should_return_standing_map_for_terrible(self):
        # given
        self.details.standing_level = PocoDetails.StandingLevel.TERRIBLE
        self.details.allow_access_with_standings = True
        # when
        result = self.details.standing_level_access_map()
        # then
        self.assertDictEqual(
            result,
            {
                "NONE": False,
                "TERRIBLE": True,
                "BAD": True,
                "NEUTRAL": True,
                "GOOD": True,
                "EXCELLENT": True,
            },
        )



import datetime as dt
from typing import List, Optional

import factory
import factory.fuzzy
import pytz
import yaml

from django.db.models import Max
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveMoon, EvePlanet, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    UserMainFactory,
)

from structures.core.notification_types import NotificationType
from structures.models import (
    FuelAlertConfig,
    GeneratedNotification,
    JumpFuelAlertConfig,
    Notification,
    Owner,
    OwnerCharacter,
    Structure,
    StructureTag,
    Webhook,
)

# eve universe (within structures)


def datetime_to_esi(my_dt: dt.datetime) -> str:
    """Convert datetime to ESI datetime string."""
    utc_dt = my_dt.astimezone(pytz.utc)
    return utc_dt.strftime("%Y-%m-%dT%H:%M:%SZ")


class EveEntityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EveEntity
        django_get_or_create = ("id", "name")

    category = EveEntity.CATEGORY_CHARACTER

    @factory.lazy_attribute
    def id(self):
        if self.category == EveEntity.CATEGORY_CHARACTER:
            obj = EveCharacterFactory()
            return obj.character_id
        if self.category == EveEntity.CATEGORY_CORPORATION:
            obj = EveCorporationInfoFactory()
            return obj.corporation_id
        if self.category == EveEntity.CATEGORY_ALLIANCE:
            obj = EveAllianceInfoFactory()
            return obj.alliance_id
        raise NotImplementedError(f"Unknown category: {self.category}")


class EveEntityCharacterFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"character_name_{n}")
    category = EveEntity.CATEGORY_CHARACTER


class EveEntityCorporationFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"corporation_name_{n}")
    category = EveEntity.CATEGORY_CORPORATION


class EveEntityAllianceFactory(EveEntityFactory):
    name = factory.Sequence(lambda n: f"alliance_name_{n}")
    category = EveEntity.CATEGORY_ALLIANCE


# Structures objects


class FuelAlertConfigFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = FuelAlertConfig

    start = 48
    end = 0
    repeat = 12


class JumpFuelAlertConfigFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = JumpFuelAlertConfig

    threshold = 100


class UserMainDefaultFactory(UserMainFactory):
    """Default user in Structures."""

    main_character__scopes = Owner.get_esi_scopes()
    permissions__ = [
        "structures.basic_access",
        "structures.view_corporation_structures",
    ]


class UserMainDefaultOwnerFactory(UserMainFactory):
    """Default user owning structures."""

    main_character__scopes = Owner.get_esi_scopes()
    permissions__ = [
        "structures.basic_access",
        "structures.add_structure_owner",
        "structures.view_corporation_structures",
    ]


class WebhookFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Webhook
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"name_{n}")
    url = factory.Sequence(lambda n: f"url_{n}")
    notes = factory.Faker("sentence")


class OwnerCharacterFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = OwnerCharacter

    structures_last_used_at = factory.LazyFunction(now)
    notifications_last_used_at = factory.LazyFunction(now)


class OwnerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Owner
        django_get_or_create = ("corporation",)

    assets_last_update_at = factory.LazyFunction(now)
    character_ownership = None  # no longer used
    forwarding_last_update_at = factory.LazyFunction(now)
    is_alliance_main = False
    is_up = True
    notifications_last_update_at = factory.LazyFunction(now)
    structures_last_update_at = factory.LazyFunction(now)
    corporation = factory.SubFactory(EveCorporationInfoFactory)

    @factory.post_generation
    def characters(
        obj, create: bool, extracted: Optional[List[CharacterOwnership]], **kwargs
    ):
        if not create:
            return
        if extracted:
            for character_ownership in extracted:
                OwnerCharacterFactory(
                    owner=obj, character_ownership=character_ownership
                )
            return
        character = EveCharacterFactory(corporation=obj.corporation)
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        character_ownership = user.profile.main_character.character_ownership
        OwnerCharacterFactory(owner=obj, character_ownership=character_ownership)

    @factory.post_generation
    def webhooks(obj, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for webhook in extracted:
                obj.webhooks.add(webhook)
        else:
            obj.webhooks.add(WebhookFactory())


class StructureFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Structure
        django_get_or_create = ("id",)

    fuel_expires_at = factory.LazyAttribute(lambda obj: now() + dt.timedelta(days=3))
    has_fitting = False
    has_core = False
    last_updated_at = factory.LazyFunction(now)
    name = factory.Sequence(lambda n: f"name_{n}")
    owner = factory.SubFactory(OwnerFactory)
    position_x = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    position_y = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    position_z = factory.fuzzy.FuzzyFloat(-10_000_000_000_000, 10_000_000_000_000)
    state = Structure.State.SHIELD_VULNERABLE

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Astrahus")

    @factory.lazy_attribute
    def eve_solar_system(self):
        return EveSolarSystem.objects.order_by("?").first()

    @factory.lazy_attribute
    def id(self):
        last_id = Structure.objects.aggregate(Max("id"))["id__max"] or 1_500_000_000_000
        return last_id + 1

    @factory.post_generation
    def webhooks(obj, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for webhook in extracted:
                obj.webhooks.add(webhook)


class StarbaseFactory(StructureFactory):
    has_fitting = None
    has_core = None
    state = Structure.State.POS_ONLINE

    @factory.lazy_attribute
    def eve_moon(self):
        return EveMoon.objects.order_by("?").first()

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_moon.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Caldari Control Tower")


class PocoFactory(StructureFactory):
    has_fitting = None
    has_core = None
    state = Structure.State.NA

    @factory.lazy_attribute
    def eve_planet(self):
        return EvePlanet.objects.order_by("?").first()

    @factory.lazy_attribute
    def eve_solar_system(self):
        return self.eve_planet.eve_solar_system

    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Customs Office")


class JumpGateFactory(StructureFactory):
    @factory.lazy_attribute
    def eve_type(self):
        return EveType.objects.get(name="Ansiblex Jump Gate")


class StructureTagFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = StructureTag
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"name_{n}")
    description = factory.Faker("sentence")


class NotificationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Notification
        exclude = ("text_from_dict",)

    text_from_dict = None

    created = factory.LazyFunction(now)
    is_read = False
    last_updated = factory.LazyAttribute(lambda o: o.created)
    notif_type = NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE.value
    owner = factory.SubFactory(OwnerFactory)
    sender = factory.SubFactory(EveEntityCorporationFactory, id=1000137, name="DED")
    timestamp = factory.LazyAttribute(lambda o: o.created)

    @factory.lazy_attribute
    def notification_id(self):
        last_id = (
            Notification.objects.aggregate(Max("notification_id"))[
                "notification_id__max"
            ]
            or 1_500_000_000
        )
        return last_id + 1

    @factory.lazy_attribute
    def text(self):
        if not self.text_from_dict:
            return ""
        return yaml.dump(self.text_from_dict)

    @classmethod
    def _adjust_kwargs(cls, **kwargs):
        if isinstance(kwargs["notif_type"], NotificationType):
            kwargs["notif_type"] = kwargs["notif_type"].value
        return kwargs


class GeneratedNotificationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = GeneratedNotification

    notif_type = NotificationType.TOWER_REINFORCED_EXTRA.value
    owner = factory.SubFactory(OwnerFactory)

    @factory.lazy_attribute
    def details(self):
        reinforced_until = factory.fuzzy.FuzzyDateTime(
            start_dt=now() + dt.timedelta(hours=3),
            end_dt=now() + dt.timedelta(hours=48),
        ).fuzz()
        return {"reinforced_until": reinforced_until.isoformat()}

    @factory.post_generation
    def create_structure(obj, create, extracted, **kwargs):
        """Set this param to False to disable."""
        if not create or extracted is False:
            return
        reinforced_until = dt.datetime.fromisoformat(obj.details["reinforced_until"])
        starbase = StarbaseFactory(
            owner=obj.owner,
            state=Structure.State.POS_REINFORCED,
            state_timer_end=reinforced_until,
        )
        obj.structures.add(starbase)


class RawNotificationFactory(factory.DictFactory):
    """Create a raw notification as received from ESI."""

    class Meta:
        exclude = ("data", "timestamp_dt", "sender")

    # excluded
    data = None
    timestamp_dt = None
    sender = factory.SubFactory(EveEntityCorporationFactory, id=2902, name="CONCORD")

    # included
    notification_id = factory.Sequence(lambda o: 1999000000 + o)
    type = "CorpBecameWarEligible"
    sender_id = factory.LazyAttribute(lambda o: o.sender.id)
    sender_type = factory.LazyAttribute(lambda o: o.sender.category)
    is_read = True

    @factory.lazy_attribute
    def timestamp(self):
        if not self.timestamp_dt:
            timestamp_dt = now()
        else:
            timestamp_dt = self.timestamp_dt
        return datetime_to_esi(timestamp_dt)

    @factory.lazy_attribute
    def text(self):
        if not self.data:
            return ""
        return yaml.dump(self.data)



from app_utils.django import app_labels

if "discord" in app_labels():
    import re
    from unittest.mock import patch

    from requests.exceptions import HTTPError

    from django.contrib.auth.models import Group

    from app_utils.testing import NoSocketsTestCase

    from structures.models import Notification
    from structures.tests.testdata.factories import create_webhook
    from structures.tests.testdata.helpers import (
        create_structures,
        load_entities,
        load_notification_entities,
        set_owner_character,
    )
    from structures.tests.testdata.load_eveuniverse import load_eveuniverse

    MODULE_PATH = "structures.models.notifications"

    @patch(MODULE_PATH + ".Notification._import_discord")
    @patch(MODULE_PATH + ".Webhook.send_message", spec=True)
    class TestGroupPings(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            load_eveuniverse()
            load_entities()
            cls.group_1 = Group.objects.create(name="Dummy Group 1")
            cls.group_2 = Group.objects.create(name="Dummy Group 2")
            create_structures(dont_load_entities=True)

        def setUp(self):
            _, self.owner = set_owner_character(character_id=1001)
            load_notification_entities(self.owner)

        @staticmethod
        def _my_group_to_role(group: Group) -> dict:
            if not isinstance(group, Group):
                raise TypeError("group must be of type Group")
            return {"id": group.pk, "name": group.name}

        def test_can_ping_via_webhook(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = create_webhook()
            webhook.ping_groups.add(self.group_1)
            obj = Notification.objects.get(notification_id=1000000509)
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_1.pk}>", kwargs["content"])

        def test_can_ping_via_owner(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = create_webhook()
            self.owner.ping_groups.add(self.group_2)
            obj = Notification.objects.get(notification_id=1000000509)
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_2.pk}>", kwargs["content"])

        def test_can_ping_both(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = create_webhook()
            webhook.ping_groups.add(self.group_1)
            self.owner.ping_groups.add(self.group_2)
            obj = Notification.objects.get(notification_id=1000000509)
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertIn(f"<@&{self.group_1.pk}>", kwargs["content"])
            self.assertIn(f"<@&{self.group_2.pk}>", kwargs["content"])

        def test_no_ping_if_not_set(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            webhook = create_webhook()
            obj = Notification.objects.get(notification_id=1000000509)
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertFalse(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertFalse(re.search(r"(<@&\d+>)", kwargs["content"]))

        def test_can_handle_http_error(self, mock_send_message, mock_import_discord):
            # given
            mock_send_message.return_value = 1
            mock_import_discord.return_value.objects.group_to_role.side_effect = (
                HTTPError
            )
            webhook = create_webhook()
            webhook.ping_groups.add(self.group_1)
            obj = Notification.objects.get(notification_id=1000000509)
            # when
            result = obj.send_to_webhook(webhook)
            # then
            self.assertTrue(result)
            self.assertTrue(mock_import_discord.called)
            _, kwargs = mock_send_message.call_args
            self.assertFalse(re.search(r"(<@&\d+>)", kwargs["content"]))
